@@ -3,19 +3,26 @@ package Ora2Pg::MySQL;
 use vars qw($VERSION);
 use strict;
 
+use DBI;
 use POSIX qw(locale_h);
 
 #set locale to LC_NUMERIC C
 setlocale(LC_NUMERIC,"C");
 
 
-$VERSION = '23.0';
+$VERSION = '23.2';
 
 # Some function might be excluded from export and assessment.
 our @EXCLUDED_FUNCTION = ('SQUIRREL_GET_ERROR_OFFSET');
 
-# These definitions can be overriden from configuration file
-our %MYSQL_TYPE = (
+# These definitions can be overriden from configuration
+# file using the DATA_TYPË configuration directive.
+our %SQL_TYPE = (
+ 	'TINYINT UNSIGNED' => 'smallint',
+ 	'SMALLINT UNSIGNED' => 'integer',
+ 	'MEDIUMINT UNSIGNED' => 'integer',
+ 	'BIGINT UNSIGNED' => 'numeric',
+ 	'INT UNSIGNED' => 'bigint',
 	'TINYINT' => 'smallint', # 1 byte
 	'SMALLINT' => 'smallint', # 2 bytes
 	'MEDIUMINT' => 'integer', # 3 bytes
@@ -55,6 +62,75 @@ our %MYSQL_TYPE = (
 	'UNSIGNED' => 'bigint'
 );
 
+sub _db_connection
+{
+	my $self = shift;
+
+	$self->logit("Trying to connect to database: $self->{oracle_dsn}\n", 1) if (!$self->{quiet});
+
+	if (!defined $self->{oracle_pwd})
+	{
+		eval("use Term::ReadKey;");
+		if (!$@) {
+			$self->{oracle_user} = $self->_ask_username('MySQL') unless (defined $self->{oracle_user});
+			$self->{oracle_pwd} = $self->_ask_password('MySQL');
+		}
+	}
+
+	my $dbh = DBI->connect("$self->{oracle_dsn}", $self->{oracle_user}, $self->{oracle_pwd}, {
+			'RaiseError' => 1,
+			AutoInactiveDestroy => 1,
+			mysql_enable_utf8 => 1,
+			mysql_conn_attrs => { program_name => 'ora2pg ' || $VERSION }
+		}
+	);
+
+	# Check for connection failure
+	if (!$dbh) {
+		$self->logit("FATAL: $DBI::err ... $DBI::errstr\n", 0, 1);
+	}
+
+	# Use consistent reads for concurrent dumping...
+	#$dbh->do('START TRANSACTION WITH CONSISTENT SNAPSHOT;') || $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+	if ($self->{debug} && !$self->{quiet}) {
+		$self->logit("Isolation level: $self->{transaction}\n", 1);
+	}
+	my $sth = $dbh->prepare($self->{transaction}) or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+	$sth->execute or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+	$sth->finish;
+
+	# Get SQL_MODE from the MySQL database
+	$sth = $dbh->prepare('SELECT @@sql_mode') or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+	$sth->execute or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+	while (my $row = $sth->fetch) {
+		$self->{mysql_mode} = $row->[0];
+	}
+	$sth->finish;
+
+	if ($self->{nls_lang})
+	{
+		if ($self->{debug} && !$self->{quiet}) {
+			$self->logit("Set default encoding to '$self->{nls_lang}' and collate to '$self->{nls_nchar}'\n", 1);
+		}
+		my $collate = '';
+		$collate = " COLLATE '$self->{nls_nchar}'" if ($self->{nls_nchar});
+		$sth = $dbh->prepare("SET NAMES '$self->{nls_lang}'$collate") or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+		$sth->execute or $self->logit("FATAL: " . $dbh->errstr . "\n", 0, 1);
+		$sth->finish;
+	}
+	# Force execution of initial command
+	$self->_ora_initial_command($dbh);
+
+	if ($self->{mysql_mode} =~ /PIPES_AS_CONCAT/) {
+		$self->{mysql_pipes_as_concat} = 1;
+	}
+
+	# Instruct Ora2Pg that the database engine is mysql
+	$self->{is_mysql} = 1;
+
+	return $dbh;
+}
+
 sub _get_version
 {
 	my $self = shift;
@@ -165,7 +241,6 @@ Returns a handle to a DB query statement.
 sub _table_info
 {
 	my $self = shift;
-	my $do_real_row_count = shift;
 
 	# First register all tablespace/table in memory from this database
 	my %tbspname = ();
@@ -201,8 +276,9 @@ sub _table_info
 
 	my %tables_infos = ();
 	my %comments = ();
-	my $sql = "SELECT TABLE_NAME,TABLE_COMMENT,TABLE_TYPE,TABLE_ROWS,ROUND( ( data_length + index_length) / 1024 / 1024, 2 ) AS \"Total Size Mb\", AUTO_INCREMENT, ENGINE FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='BASE TABLE' AND TABLE_SCHEMA = '$self->{schema}'";
+	my $sql = "SELECT TABLE_NAME,TABLE_COMMENT,TABLE_TYPE,TABLE_ROWS,ROUND( ( data_length + index_length) / 1024 / 1024, 2 ) AS \"Total Size Mb\", AUTO_INCREMENT, CREATE_OPTIONS FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE='BASE TABLE' AND TABLE_SCHEMA = '$self->{schema}'";
 	$sql .= $self->limit_to_objects('TABLE', 'TABLE_NAME');
+
 	$sth = $self->{dbh}->prepare( $sql ) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 	while (my $row = $sth->fetch)
@@ -219,6 +295,7 @@ sub _table_info
 		$tables_infos{$row->[0]}{tablespace} = 0;
 		$tables_infos{$row->[0]}{auto_increment} = $row->[5] || 0;
 		$tables_infos{$row->[0]}{tablespace} = $tbspname{$row->[0]} || '';
+		$tables_infos{$row->[0]}{partitioned} = ($row->[6] eq 'partitioned' || exists $self->{partitions}{$row->[0]}) ? 1 : 0;
 
 		# Get creation option unavailable in information_schema
 		if ($row->[6] eq 'FEDERATED')
@@ -234,16 +311,6 @@ sub _table_info
 			}
 			$sth2->finish();
 		}
-		if ($do_real_row_count)
-		{
-			$self->logit("DEBUG: looking for real row count for table $row->[0] (aka using count(*))...\n", 1);
-			$sql = "SELECT COUNT(*) FROM `$row->[0]`";
-			my $sth2 = $self->{dbh}->prepare( $sql ) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
-			$sth2->execute or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
-			my $size = $sth2->fetch();
-			$sth2->finish();
-			$tables_infos{$row->[0]}{num_rows} = $size->[0];
-		}
 	}
 	$sth->finish();
 
@@ -314,8 +381,9 @@ sub _column_info
 	# EXTRA                    | varchar(27)         | NO   |     |         |       |
 	# PRIVILEGES               | varchar(80)         | NO   |     |         |       |
 	# COLUMN_COMMENT           | varchar(1024)       | NO   |     |         |       |
+	# GENERATION_EXPRESSION    | longtext            | NO   |     | NULL    |       |
 
-	my $str = qq{SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH, IS_NULLABLE, COLUMN_DEFAULT, NUMERIC_PRECISION, NUMERIC_SCALE, CHARACTER_OCTET_LENGTH, TABLE_NAME, '' AS OWNER, '' AS VIRTUAL_COLUMN, ORDINAL_POSITION, EXTRA, COLUMN_TYPE
+	my $str = qq{SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH, IS_NULLABLE, COLUMN_DEFAULT, NUMERIC_PRECISION, NUMERIC_SCALE, CHARACTER_OCTET_LENGTH, TABLE_NAME, '' AS OWNER, '' AS VIRTUAL_COLUMN, ORDINAL_POSITION, EXTRA, COLUMN_TYPE, GENERATION_EXPRESSION
 FROM INFORMATION_SCHEMA.COLUMNS
 $condition
 ORDER BY ORDINAL_POSITION};
@@ -330,15 +398,58 @@ ORDER BY ORDINAL_POSITION};
 	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 
 	# Expected columns information stored in hash 
+	# COLUMN_NAME,DATA_TYPE,DATA_LENGTH,NULLABLE,DATA_DEFAULT,DATA_PRECISION,DATA_SCALE,CHAR_LENGTH,TABLE_NAME,OWNER,VIRTUAL_COLUMN,POSITION,AUTO_INCREMENT,SRID,SDO_DIM,SDO_GTYPE
 	# COLUMN_NAME,DATA_TYPE,DATA_LENGTH,NULLABLE,DATA_DEFAULT,DATA_PRECISION,DATA_SCALE,CHAR_LENGTH,TABLE_NAME,OWNER,VIRTUAL_COLUMN,POSITION,AUTO_INCREMENT,ENUM_INFO
 	my %data = ();
 	my $pos = 0;
 	while (my $row = $sth->fetch)
 	{
 		if ($row->[1] eq 'enum') {
-			$row->[1] = $row->[-1];
+			$row->[1] = $row->[-2];
 		}
+		if ($row->[13] =~ /unsigned/) {
+			$row->[1] .= ' unsigned';
+		}
+
 		$row->[10] = $pos;
+		$row->[12] =~ s/\s+ENABLE//is;
+		if ($row->[12] =~ s/\bGENERATED\s+(ALWAYS|BY\s+DEFAULT)\s+(ON\s+NULL\s+)?AS\s+IDENTITY\s*(.*)//is)
+		{
+			$self->{identity_info}{$row->[8]}{$row->[0]}{generation} = $1;
+			my $options = $3;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} = $3;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/(SCALE|EXTEND|SESSION)_FLAG: .//isg;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/KEEP_VALUE: .//is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/(START WITH):/$1/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/(INCREMENT BY):/$1/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/MAX_VALUE:/MAXVALUE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/MIN_VALUE:/MINVALUE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/CYCLE_FLAG: N/NO CYCLE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/NOCYCLE/NO CYCLE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/CYCLE_FLAG: Y/CYCLE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/CACHE_SIZE:/CACHE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/CACHE_SIZE:/CACHE/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/ORDER_FLAG: .//is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/,//gs;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s$//s;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/CACHE\s+0/CACHE 1/is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s*NOORDER//is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s*NOKEEP//is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s*NOSCALE//is;
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s*NOT\s+NULL//is;
+			# Be sure that we don't exceed the bigint max value,
+			# we assume that the increment is always positive
+			if ($self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ /MAXVALUE\s+(\d+)/is) {
+				$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/(MAXVALUE)\s+\d+/$1 9223372036854775807/is;
+			}
+			$self->{identity_info}{$row->[8]}{$row->[0]}{options} =~ s/\s+/ /igs;
+		}
+		elsif ($row->[12] =~ s/\bGENERATED\b//is)
+		{
+			$row->[10] = 'YES';
+			$row->[14] =~ s/\`//g;
+			$row->[4] = $row->[14];
+		}
 		push(@{$data{"$row->[8]"}{"$row->[0]"}}, @$row);
 		pop(@{$data{"$row->[8]"}{"$row->[0]"}});
 		$pos++;
@@ -349,7 +460,7 @@ ORDER BY ORDINAL_POSITION};
 
 sub _get_indexes
 {
-	my ($self, $table, $owner) = @_;
+	my ($self, $table, $owner, $generated_indexes) = @_;
 
 	my $condition = '';
 	$condition = " FROM $self->{schema}" if ($self->{schema});
@@ -407,15 +518,23 @@ sub _get_indexes
 			# Enclose with double quote if required
 			$row->[4] = $self->quote_object_name($row->[4]);
 
-			if ($self->{preserve_case}) {
+			if ($self->{preserve_case})
+			{
 				if (($row->[4] !~ /".*"/) && ($row->[4] !~ /\(.*\)/)) {
 					$row->[4] =~ s/^/"/;
 					$row->[4] =~ s/$/"/;
 				}
 			}
+			# Set the index expression
+			if ($row->[14] ne '') {
+				$row->[4] = $row->[14];
+			}
+			# Append DESC sort order when not default to ASC
+			if ($row->[5] eq 'D') {
+				$row->[4] .= " DESC";
+			}
 			push(@{$data{$row->[0]}{$idxname}}, $row->[4]);
 			$index_tablespace{$row->[0]}{$idxname} = '';
-
 		}
 	}
 
@@ -691,6 +810,65 @@ sub _unique_key
 	return %result;
 }
 
+sub _check_constraint
+{
+	my ($self, $table, $owner) = @_;
+
+	if ($self->{db_version} < '8.0.0') {
+		return;
+	}
+
+	my $condition = '';
+	$condition .= "AND TABLE_NAME='$table' " if ($table);
+	$condition .= $self->limit_to_objects('CKEY|TABLE', 'CONSTRAINT_NAME|TABLE_NAME');
+
+	my $sql = qq{SELECT CONSTRAINT_NAME, TABLE_NAME FROM information_schema.TABLE_CONSTRAINTS WHERE CONSTRAINT_TYPE = 'CHECK' AND TABLE_SCHEMA = '$self->{schema}' $condition};
+	my $sth = $self->{dbh}->prepare($sql) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+
+	my %data = ();
+	while (my $row = $sth->fetch)
+	{
+		# Pour chaque retour SHOW CREATE TABLE xxxx;
+		my $sql2 = "SHOW CREATE TABLE $row->[1];";
+		my $sth2 = $self->{dbh}->prepare($sql2) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+		$sth2->execute() or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+		# Parsing de   CONSTRAINT `CHK_CONSTR` CHECK (((`Age` >= 18) and (`City` = _utf8mb4'Bangalore')))
+		while (my $r = $sth2->fetch)
+		{
+			$r->[1] =~ s/`//g;
+			my @def = split(/[\r\n]+/, $r->[1]);
+			foreach my $l (@def)
+			{
+				if ($l =~ s/.*CONSTRAINT $row->[0] CHECK (.*)/$1/)
+				{
+					$l =~ s/(LIKE|=) _[^']+('[^']+')/$1 $2/g;
+					$data{$row->[1]}{constraint}{$row->[0]}{condition} = $l;
+					$data{$row->[1]}{constraint}{$row->[0]}{validate}  = 'Y';
+				}
+			}
+		}
+	}
+
+	return %data;
+}
+
+sub _get_external_tables
+{
+	my ($self) = @_;
+
+	# There is no external table in MySQL
+	return;
+}
+
+sub _get_directory
+{
+	my ($self) = @_;
+
+	# There is no external table in MySQL
+	return;
+}
+
 sub _get_functions
 {
 	my $self = shift;
@@ -788,6 +966,12 @@ sub _lookup_function
 
         # Split data into declarative and code part
         ($fct_detail{declare}, $fct_detail{code}) = split(/\bBEGIN\b/i, $code, 2);
+	if (!$fct_detail{code})
+	{
+		if ($fct_detail{declare} =~ s/(RETURN .*)$//i) {
+			$fct_detail{code} = "BEGIN\n    $1;\nEND\n";
+		}
+	}
 	return if (!$fct_detail{code});
 
 	# Remove any label that was before the main BEGIN block
@@ -796,7 +980,8 @@ sub _lookup_function
         @{$fct_detail{param_types}} = ();
 
         if ( ($fct_detail{declare} =~ s/(.*?)\b(FUNCTION|PROCEDURE)\s+([^\s\(]+)\s*(\(.*\))\s+RETURNS\s+(.*)//is) ||
-        ($fct_detail{declare} =~ s/(.*?)\b(FUNCTION|PROCEDURE)\s+([^\s\(]+)\s*(\(.*\))//is) ) {
+			($fct_detail{declare} =~ s/(.*?)\b(FUNCTION|PROCEDURE)\s+([^\s\(]+)\s*(\(.*\))//is) )
+	{
                 $fct_detail{before} = $1;
                 $fct_detail{type} = uc($2);
                 $fct_detail{name} = $3;
@@ -876,6 +1061,9 @@ sub _lookup_function
 
 	($fct_detail{code}, $fct_detail{declare}) = replace_mysql_variables($self, $fct_detail{code}, $fct_detail{declare});
 
+	# Remove %ROWTYPE from return type
+	$fct_detail{func_ret_type} =~ s/\%ROWTYPE//igs;
+
 	return %fct_detail;
 }
 
@@ -998,11 +1186,9 @@ sub _list_all_funtions
 	return @functions;
 }
 
-
-
 sub _sql_type
 {
-        my ($self, $type, $len, $precision, $scale) = @_;
+        my ($self, $type, $len, $precision, $scale, $default, $no_blob_to_oid) = @_;
 
 	my $data_type = '';
 
@@ -1026,26 +1212,47 @@ sub _sql_type
 	}
 
         # Override the length
-        $len = $precision if ( ((uc($type) eq 'NUMBER') || (uc($type) eq 'BIT')) && $precision );
-        if (exists $self->{data_type}{uc($type)}) {
+	#$len = $precision if ( ((uc($type) eq 'NUMBER') || (uc($type) eq 'BIT')) && $precision );
+        $len = $precision if ($precision);
+        if (exists $self->{data_type}{uc($type)})
+	{
 		$type = uc($type); # Force uppercase
-		if ($len) {
-			if ( ($type eq "CHAR") || ($type =~ /VARCHAR/) ) {
+		if ($len)
+		{
+			if ( ($type eq "CHAR") || ($type =~ /VARCHAR/) )
+			{
 				# Type CHAR have default length set to 1
 				# Type VARCHAR(2) must have a specified length
 				$len = 1 if (!$len && ($type eq "CHAR"));
                 		return "$self->{data_type}{$type}($len)";
-			} elsif ($type eq 'BIT') {
+			}
+			elsif ($type eq 'BIT')
+			{
 				if ($precision) {
 					return "$self->{data_type}{$type}($precision)";
 				} else {
 					return $self->{data_type}{$type};
 				}
-			} elsif ($type =~ /(TINYINT|SMALLINT|MEDIUMINT|INTEGER|BIGINT|INT|REAL|DOUBLE|FLOAT|DECIMAL|NUMERIC)/i) {
+			}
+		       	elsif ($type =~ /(TINYINT|SMALLINT|MEDIUMINT|INTEGER|BIGINT|INT|REAL|DOUBLE|FLOAT|DECIMAL|NUMERIC)/i)
+			{
 				# This is an integer
-				if (!$scale) {
-					if ($precision) {
-						if ($self->{pg_integer_type}) {
+				if (!$scale)
+				{
+					if ($type =~ /UNSIGNED/&& $precision)
+					{
+						# Replace MySQL type UNSIGNED in cast
+						$type =~ s/TINYINT UNSIGNED/smallint/igs;
+						$type =~ s/SMALLINT UNSIGNED/integer/igs;
+						$type =~ s/MEDIUMINT UNSIGNED/integer/igs;
+						$type =~ s/BIGINT UNSIGNED/numeric($precision)/igs;
+						$type =~ s/INT UNSIGNED/bigint/igs;
+						return $type;
+					}
+					elsif ($precision)
+					{
+						if ($self->{pg_integer_type})
+						{
 							if ($precision < 5) {
 								return 'smallint';
 							} elsif ($precision <= 9) {
@@ -1055,13 +1262,19 @@ sub _sql_type
 							}
 						}
 						return "numeric($precision)";
-					} else {
+					}
+					else
+					{
 						# Most of the time interger should be enought?
 						return $self->{data_type}{$type};
 					}
-				} else {
-					if ($precision) {
-						if ($type !~ /DOUBLE/ && $self->{pg_numeric_type}) {
+				}
+				else
+				{
+					if ($precision)
+					{
+						if ($type !~ /(DOUBLE|DECIMAL)/ && $self->{pg_numeric_type})
+						{
 							if ($precision <= 6) {
 								return 'real';
 							} else {
@@ -1073,7 +1286,9 @@ sub _sql_type
 				}
 			}
 			return $self->{data_type}{$type};
-		} else {
+		}
+		else
+		{
 			return $self->{data_type}{$type};
 		}
         }
@@ -1103,11 +1318,13 @@ sub replace_sql_type
 		$mysqltype_regex .= quotemeta($_) . '|';
 	}
 	$mysqltype_regex =~ s/\|$//;
-	while ($str =~ /(.*)\b($mysqltype_regex)\s*\(([^\)]+)\)/i) {
+	while ($str =~ /(.*)\b($mysqltype_regex)\s*\(([^\)]+)\)/i)
+	{
 		my $backstr = $1;
 		my $type = uc($2);
 		my $args = $3;
-		if (uc($type) eq 'ENUM') {
+		if (uc($type) eq 'ENUM')
+		{
 			# Prevent from infinit loop
 			$str =~ s/\(/\%\|/s;
 			$str =~ s/\)/\%\|\%/s;
@@ -1167,10 +1384,12 @@ sub replace_sql_type
 	# Replace datatype even without precision
 	my %recover_type = ();
 	my $i = 0;
-	foreach my $type (sort { length($b) <=> length($a) } keys %data_type) {
+	foreach my $type (sort { length($b) <=> length($a) } keys %data_type)
+	{
 		# Keep enum as declared, we are not in table definition
 		next if (uc($type) eq 'ENUM');
-		while ($str =~ s/\b$type\b/%%RECOVER_TYPE$i%%/is) {
+		while ($str =~ s/\b$type\b/%%RECOVER_TYPE$i%%/is)
+		{
 			$recover_type{$i} = $data_type{$type};
 			$i++;
 		}
@@ -1252,7 +1471,7 @@ sub _get_partitions
 	my $str = qq{
 SELECT TABLE_NAME, PARTITION_ORDINAL_POSITION, PARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, PARTITION_METHOD, PARTITION_EXPRESSION
 FROM INFORMATION_SCHEMA.PARTITIONS
-WHERE PARTITION_NAME IS NOT NULL AND SUBPARTITION_NAME IS NULL AND (PARTITION_METHOD = 'RANGE' OR PARTITION_METHOD = 'LIST')
+WHERE PARTITION_NAME IS NOT NULL
 };
 	$str .= $self->limit_to_objects('TABLE|PARTITION', 'TABLE_NAME|PARTITION_NAME');
 	if ($self->{schema}) {
@@ -1264,17 +1483,34 @@ WHERE PARTITION_NAME IS NOT NULL AND SUBPARTITION_NAME IS NULL AND (PARTITION_ME
 	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 	my %parts = ();
 	my %default = ();
-	while (my $row = $sth->fetch) {
-		if ( ($row->[3] eq 'MAXVALUE') || ($row->[3] eq 'DEFAULT')) {
+	while (my $row = $sth->fetch)
+	{
+		if ($row->[3] =~ /^MAXVALUE(?:,MAXVALUE)*$/ || $row->[3] eq 'DEFAULT')
+		{
 			$default{$row->[0]} = $row->[2];
 			next;
 		}
-		my $i = $#{$parts{$row->[0]}{$row->[1]}{$row->[2]}} + 1;
-		push(@{$parts{$row->[0]}{$row->[1]}{$row->[2]}}, { 'type' => $row->[5], 'value' => $row->[3], 'column' => $row->[6], 'colpos' => $i, 'tablespace' => $row->[4], 'owner' => ''});
-		$self->logit(".",1);
+		$parts{$row->[0]}{$row->[1]}{name} = $row->[2];
+		$row->[6] =~ s/\`//g;
+		$row->[3] =~ s/\`//g;
+
+		$row->[5] = 'HASH' if ($row->[5] =~ /KEY/);
+
+		if ($row->[5] =~ s/ COLUMNS//)
+		{
+			my $i = 0;
+			foreach my $c (split(',', $row->[6]))
+			{
+				push(@{$parts{$row->[0]}{$row->[1]}{info}}, { 'type' => $row->[5], 'value' => $row->[3], 'column' => $c, 'colpos' => $i, 'tablespace' => $row->[4], 'owner' => ''});
+				$i++;
+			}
+		}
+		else
+		{
+			@{$parts{$row->[0]}{$row->[1]}{info}} = ( { 'type' => $row->[5], 'value' => $row->[3], 'expression' => $row->[6], 'colpos' => 0, 'tablespace' => $row->[4], 'owner' => '' } );
+		}
 	}
 	$sth->finish;
-	$self->logit("\n", 1);
 
 	return \%parts, \%default;
 }
@@ -1291,42 +1527,51 @@ sub _get_subpartitions
 
 	# Retrieve all partitions.
 	my $str = qq{
-SELECT TABLE_NAME, SUBPARTITION_ORDINAL_POSITION, SUBPARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, SUBPARTITION_METHOD, SUBPARTITION_EXPRESSION
+SELECT TABLE_NAME, SUBPARTITION_ORDINAL_POSITION, SUBPARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, SUBPARTITION_METHOD, SUBPARTITION_EXPRESSION,PARTITION_NAME
 FROM INFORMATION_SCHEMA.PARTITIONS
-WHERE SUBPARTITION_NAME IS NOT NULL AND SUBPARTITION_EXPRESSION IS NOT NULL AND (SUBPARTITION_METHOD = 'RANGE' OR SUBPARTITION_METHOD = 'LIST')
+WHERE SUBPARTITION_NAME IS NOT NULL AND SUBPARTITION_EXPRESSION IS NOT NULL
 };
 	$str .= $self->limit_to_objects('TABLE|PARTITION', 'TABLE_NAME|PARTITION_NAME');
 	if ($self->{schema}) {
 		$str .= " AND TABLE_SCHEMA ='$self->{schema}'\n";
 	}
-	$str .= " ORDER BY TABLE_NAME,PARTITION_ORDINAL_POSITION\n";
+	$str .= " ORDER BY TABLE_NAME,PARTITION_ORDINAL_POSITION,SUBPARTITION_ORDINAL_POSITION\n";
 	my $sth = $self->{dbh}->prepare($str) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 	my %subparts = ();
 	my %default = ();
-	while (my $row = $sth->fetch) {
-		if ( ($row->[3] eq 'MAXVALUE') || ($row->[3] eq 'DEFAULT')) {
+	while (my $row = $sth->fetch)
+	{
+		if ($row->[3] =~ /^MAXVALUE(?:,MAXVALUE)*$/ || $row->[3] eq 'DEFAULT')
+		{
 			$default{$row->[0]} = $row->[2];
 			next;
 		}
-		my $i = $#{$subparts{$row->[0]}{$row->[1]}{$row->[2]}} + 1;
-		push(@{$subparts{$row->[0]}{$row->[1]}{$row->[2]}}, { 'type' => $row->[5], 'value' => $row->[3], 'column' => $row->[6], 'colpos' => $i, 'tablespace' => $row->[4], 'owner' => ''});
-		$self->logit(".",1);
+		$subparts{$row->[0]}{$row->[7]}{$row->[1]}{name} = $row->[2];
+		my $i = 0;
+		$row->[6] =~ s/\`//g;
+		$row->[3] =~ s/\`//g;
+		$row->[5] =~ s/ COLUMNS//;
+		$row->[5] = 'HASH' if ($row->[5] =~ /KEY/);
+		foreach my $c (split(',', $row->[6]))
+		{
+			push(@{$subparts{$row->[0]}{$row->[7]}{$row->[1]}{info}}, { 'type' => $row->[5], 'value' => $row->[3], 'column' => $c, 'colpos' => $i, 'tablespace' => $row->[4], 'owner' => ''});
+			$i++;
+		}
 	}
 	$sth->finish;
-	$self->logit("\n", 1);
 
 	return \%subparts, \%default;
 }
 
-=head2 _get_partitions_list
+=head2 _get_partitions_type
 
 This function implements a MySQL-native partitions information.
 Return a hash of the partition table_name => type
 
 =cut
 
-sub _get_partitions_list
+sub _get_partitions_type
 {
 	my($self) = @_;
 
@@ -1361,12 +1606,13 @@ Return a hash of the partitioned table with the number of partition
 
 sub _get_partitioned_table
 {
-	my($self) = @_;
+	my ($self, %subpart) = @_;
 
 	# Retrieve all partitions.
 	my $str = qq{
-SELECT TABLE_NAME, PARTITION_ORDINAL_POSITION, PARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, PARTITION_METHOD
-FROM INFORMATION_SCHEMA.PARTITIONS WHERE SUBPARTITION_NAME IS NULL AND PARTITION_NAME IS NOT NULL
+SELECT TABLE_NAME, PARTITION_METHOD, PARTITION_ORDINAL_POSITION, PARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, PARTITION_EXPRESSION
+FROM INFORMATION_SCHEMA.PARTITIONS WHERE PARTITION_NAME IS NOT NULL
+     AND (PARTITION_METHOD LIKE 'RANGE%' OR PARTITION_METHOD LIKE 'LIST%' OR PARTITION_METHOD LIKE 'HASH%' OR PARTITION_METHOD LIKE 'KEY%' OR PARTITION_METHOD LIKE 'LINEAR KEY%')
 };
 	$str .= $self->limit_to_objects('TABLE|PARTITION','TABLE_NAME|PARTITION_NAME');
 	if ($self->{schema}) {
@@ -1378,8 +1624,70 @@ FROM INFORMATION_SCHEMA.PARTITIONS WHERE SUBPARTITION_NAME IS NULL AND PARTITION
 	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
 
 	my %parts = ();
-	while (my $row = $sth->fetch) {
-		$parts{$row->[0]}++ if ($row->[2]);
+	while (my $row = $sth->fetch)
+	{
+		$parts{"\L$row->[0]\E"}{count} = 0;
+		$parts{"\L$row->[0]\E"}{composite} = 0;
+		if (exists $subpart{"\L$row->[0]\E"})
+		{
+			$parts{"\L$row->[0]\E"}{composite} = 1;
+			foreach my $k (keys %{$subpart{"\L$row->[0]\E"}}) {
+				$parts{"\L$row->[0]\E"}{count} += $subpart{"\L$row->[0]\E"}{$k}{count};
+			}
+			$parts{"\L$row->[0]\E"}{count}++;
+		} else {
+			$parts{"\L$row->[0]\E"}{count}++;
+		}
+		$parts{"\L$row->[0]\E"}{type} = $row->[1];
+		$row->[6] =~ s/\`//g;
+
+		if ($row->[1] =~ /KEY/)
+		{
+			$parts{"\L$row->[0]\E"}{type} = 'HASH';
+			my $sql = "SHOW INDEX FROM `$row->[0]`";
+			my $sth2 = $self->{dbh}->prepare($sql) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+			$sth2->execute() or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+			my @ucol = ();
+			while (my $r = $sth2->fetch)
+			{
+				#Table : The name of the table.
+				#Non_unique : 0 if the index cannot contain duplicates, 1 if it can.
+				#Key_name : The name of the index. If the index is the primary key, the name is always PRIMARY.
+				#Seq_in_index : The column sequence number in the index, starting with 1.
+				#Column_name : The column name.
+				#Collation : How the column is sorted in the index. In MySQL, this can have values “A” (Ascending) or NULL (Not sorted).
+				#Cardinality : An estimate of the number of unique values in the index.
+				#Sub_part : The number of indexed characters if the column is only partly indexed, NULL if the entire column is indexed.
+				#Packed : Indicates how the key is packed. NULL if it is not.
+				#Null : Contains YES if the column may contain NULL values and '' if not.
+				#Index_type : The index method used (BTREE, FULLTEXT, HASH, RTREE).
+				#Comment : Information about the index not described in its own column, such as disabled if the index is disabled. 
+				if ($r->[2] eq 'PRIMARY') {
+					push(@{ $parts{"\L$row->[0]\E"}{columns} }, $r->[4]) if (!grep(/^$r->[4]$/, @{ $parts{"\L$row->[0]\E"}{columns} }));
+				} elsif (!$r->[1]) {
+					push(@ucol, $r->[4]) if (!grep(/^$r->[4]$/, @ucol));
+				}
+			}
+			$sth2->finish;
+			if ($#{ $parts{"\L$row->[0]\E"}{columns} } < 0) {
+				if ($#ucol >= 0) {
+					push(@{ $parts{"\L$row->[0]\E"}{columns} }, @ucol);
+				} else {
+					$row->[6] =~ s/[\(\)\s]//g;
+					@{ $parts{"\L$row->[0]\E"}{columns} } = split(',', $row->[6]);
+				}
+			}
+
+		}
+		if ($parts{"\L$row->[0]\E"}{type} =~ s/ COLUMNS//)
+		{
+			$row->[6] =~ s/[\(\)\s]//g;
+			@{ $parts{"\L$row->[0]\E"}{columns} } = split(',', $row->[6]);
+		}
+		else
+		{
+			$parts{"\L$row->[0]\E"}{expression} = $row->[6];
+		}
 	}
 	$sth->finish;
 
@@ -1455,7 +1763,8 @@ sub _get_objects
 	my $str = qq{
 SELECT TABLE_NAME||'_'||PARTITION_NAME
 FROM INFORMATION_SCHEMA.PARTITIONS
-WHERE SUBPARTITION_NAME IS NULL AND (PARTITION_METHOD = 'RANGE' OR PARTITION_METHOD = 'LIST')
+WHERE SUBPARTITION_NAME IS NULL
+     AND (SUBPARTITION_METHOD LIKE 'RANGE%' OR SUBPARTITION_METHOD LIKE 'LIST%' OR SUBPARTITION_METHOD LIKE 'HASH%' OR SUBPARTITION_METHOD LIKE 'KEY%' OR SUBPARTITION_METHOD LIKE 'LINEAR KEY%')
 };
 	$sql .= $self->limit_to_objects('TABLE|PARTITION', 'TABLE_NAME|PARTITION_NAME');
 	if ($self->{schema}) {
@@ -1816,7 +2125,7 @@ sub _global_temp_table_info
 
 sub _encrypted_columns
 {
-        my($self) = @_;
+        my ($self, $table, $owner) = @_;
 
 	return;
 }
@@ -1825,7 +2134,74 @@ sub _get_subpartitioned_table
 {
         my($self) = @_;
 
-	return;
+	# Retrieve all partitions.
+	my $str = qq{
+SELECT TABLE_NAME, SUBPARTITION_METHOD, SUBPARTITION_ORDINAL_POSITION, PARTITION_NAME, SUBPARTITION_NAME, PARTITION_DESCRIPTION, TABLESPACE_NAME, SUBPARTITION_EXPRESSION
+FROM INFORMATION_SCHEMA.PARTITIONS WHERE SUBPARTITION_NAME IS NOT NULL
+     AND (SUBPARTITION_METHOD LIKE 'RANGE%' OR SUBPARTITION_METHOD LIKE 'LIST%' OR SUBPARTITION_METHOD LIKE 'HASH%' OR SUBPARTITION_METHOD LIKE 'KEY%' OR SUBPARTITION_METHOD LIKE 'LINEAR KEY%')
+};
+	$str .= $self->limit_to_objects('TABLE|PARTITION','TABLE_NAME|SUBPARTITION_NAME');
+	if ($self->{schema}) {
+		$str .= " AND TABLE_SCHEMA ='$self->{schema}'";
+	}
+	$str .= " ORDER BY TABLE_NAME,PARTITION_NAME,SUBPARTITION_NAME\n";
+
+	my $sth = $self->{dbh}->prepare($str) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+	$sth->execute(@{$self->{query_bind_params}}) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+
+	my %parts = ();
+	while (my $row = $sth->fetch)
+	{
+		$parts{"\L$row->[0]\E"}{"\L$row->[3]\E"}{count}++;
+		$parts{"\L$row->[0]\E"}{"\L$row->[3]\E"}{type} = $row->[1];
+		$row->[7] =~ s/\`//g;
+
+		if ($parts{"\L$row->[0]\E"}{type} =~ /KEY/)
+		{
+			$parts{"\L$row->[0]\E"}{type} = 'HASH';
+			my $sql = "SHOW INDEX FROM `$row->[0]`";
+			my $sth2 = $self->{dbh}->prepare($sql) or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+			$sth2->execute() or $self->logit("FATAL: " . $self->{dbh}->errstr . "\n", 0, 1);
+			my @ucol = ();
+			while (my $r = $sth2->fetch)
+			{
+				#Table : The name of the table.
+				#Non_unique : 0 if the index cannot contain duplicates, 1 if it can.
+				#Key_name : The name of the index. If the index is the primary key, the name is always PRIMARY.
+				#Seq_in_index : The column sequence number in the index, starting with 1.
+				#Column_name : The column name.
+				#Collation : How the column is sorted in the index. In MySQL, this can have values “A” (Ascending) or NULL (Not sorted).
+				#Cardinality : An estimate of the number of unique values in the index.
+				#Sub_part : The number of indexed characters if the column is only partly indexed, NULL if the entire column is indexed.
+				#Packed : Indicates how the key is packed. NULL if it is not.
+				#Null : Contains YES if the column may contain NULL values and '' if not.
+				#Index_type : The index method used (BTREE, FULLTEXT, HASH, RTREE).
+				#Comment : Information about the index not described in its own column, such as disabled if the index is disabled. 
+				if ($r->[2] eq 'PRIMARY') {
+					push(@{ $parts{"\L$row->[0]\E"}{columns} }, $r->[4]) if (!grep(/^$r->[4]$/, @{ $parts{"\L$row->[0]\E"}{columns} }));
+				} elsif (!$row->[1]) {
+					push(@ucol, $r->[4]) if (!grep(/^$r->[4]$/, @ucol));
+				}
+			}
+			$sth2->finish;
+			if ($#{ $parts{"\L$row->[0]\E"}{columns} } < 0) {
+				push(@{ $parts{"\L$row->[0]\E"}{columns} }, @ucol);
+			}
+		}
+
+		if ($parts{"\L$row->[0]\E"}{"\L$row->[3]\E"}{type} =~ s/ COLUMNS//)
+		{
+			$row->[7] =~ s/[\(\)\s]//g;
+			@{ $parts{"\L$row->[0]\E"}{"\L$row->[3]\E"}{columns} } = split(',', $row->[7]);
+		}
+		else
+		{
+			$parts{"\L$row->[0]\E"}{"\L$row->[3]\E"}{expression} = $row->[7];
+		}
+	}
+	$sth->finish;
+
+	return %parts;
 }
 
 # Replace IF("user_status"=0,"username",NULL)
@@ -2002,5 +2378,45 @@ sub _get_materialized_views
 	return;
 }
 
+sub _get_materialized_view_names
+{
+	my($self) = @_;
+
+	# nothing to do, materialized view are not supported by MySQL.
+	return;
+}
+
+sub _get_package_function_list
+{
+	my ($self, $owner) = @_;
+
+	# not package in MySQL
+	return;
+}
+
+sub _get_procedures
+{
+	my ($self) = @_;
+
+	# not package in MySQL
+	return _get_functions($self);
+}
+
+sub _get_types
+{
+        my ($self, $name) = @_;
+
+	# Not supported
+	return;
+}
+
+sub _col_count
+{
+        my ($self, $name) = @_;
+
+	# Not supported
+	return;
+}
+
 1;
 
